package s05.demineur;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.BufferedImage;
import java.io.File;
 
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
 
public class Demineur extends JFrame {
	private static final long serialVersionUID = 1L;
	public static final int ICON_SIZE = 50;
	public static final int CACHEE = 9;
	public static final int SELECTED = 11;
	public static final int MARQUEE = 12;
 
	private ITerrainMine terrain;
	private JButton[][] lesBoutons;
 
	public Demineur(ITerrainMine t) {
		super("Demineur");
		this.setDefaultCloseOperation(EXIT_ON_CLOSE);
		this.terrain = t;
		int nbCol = t.getNbColonnes();
		int nbLig = t.getNbLignes();
		this.getContentPane().setLayout(new GridLayout(nbLig,nbCol));
		this.lesBoutons = new JButton[nbCol][nbLig];
		for (int l=0; l<this.getTerrain().getNbLignes(); l++) {
			for (int c=0; c<this.getTerrain().getNbColonnes(); c++)  {
				this.lesBoutons[c][l] = new JButton( this.getIcone(c,l));//this.getTerrain().get(c, l).getVal()));
				this.lesBoutons[c][l].setBorder(null);
				if (this.getTerrain().get(c, l).estCachee() && !this.getTerrain().get(c, l).estMarquee()) {
					this.lesBoutons[c][l].setRolloverIcon(this.getIcone(SELECTED));
				} else {
					this.lesBoutons[c][l].setRolloverIcon(null);
				}
				this.lesBoutons[c][l].addMouseListener(new EcouteurBouton(this, this.getTerrain().get(c, l)));
				this.getContentPane().add(this.lesBoutons[c][l]);
			}
		}
		this.pack();
		this.setSize(new Dimension((int)(this.getSize().getWidth()-10), (int)(this.getSize().getHeight()-10)));
		this.setResizable(false);
		this.setVisible(true);
	}
	public JButton getBouton(int c, int l) {
		return this.lesBoutons[c][l];
	}
	public void miseAJourImages() {
		for (int l=0; l<this.getTerrain().getNbLignes(); l++) {
			for (int c=0; c<this.getTerrain().getNbColonnes(); c++)  {
 				if (this.getTerrain().get(c, l).estCachee() && !this.getTerrain().get(c, l).estMarquee()) {
					this.getBouton(c, l).setRolloverIcon(this.getIcone(SELECTED));
				} else {
					this.getBouton(c, l).setRolloverIcon(null);
				}
				this.getBouton(c, l).setIcon(this.getIcone(c,l));
			}
		}
	}
	public ITerrainMine getTerrain() {
		return this.terrain;
	}
	public ImageIcon getIcone(int col, int lig) {
		if (this.getTerrain().get(col, lig)==null) {
			System.out.println("   Aie... des cases de la grille valent null \n--> vous n'avez pas du decommenter la ligne correspondant a 1.2.2 dans TerrainMinee");
		}
		if (this.getTerrain().get(col, lig).estCachee()) {
			if (this.getTerrain().get(col, lig).estMarquee()) {
				return this.getIcone(MARQUEE);
			} else {
				return this.getIcone(CACHEE);
			}
		} else {
			return this.getIcone(this.getTerrain().get(col, lig).getValeur());
		}
	}
	public ImageIcon getIcone(int i) {
		Image img = new ImageIcon("images"+File.separator+"dalle_0"+i+".gif").getImage();
		BufferedImage buf = new BufferedImage(ICON_SIZE,ICON_SIZE,BufferedImage.TYPE_INT_ARGB);// largeur, hauteur,	BufferedImage.TYPE_INT_ARGB);
		Graphics2D g = buf.createGraphics();
		g.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
		g.drawImage(img, 0, 0, ICON_SIZE, ICON_SIZE, null);//largeur, hauteur, null);
		g.dispose();
		return (new ImageIcon(buf));
 
	}
	public static void main(String[] args) {
		String[] niveaux={"Debutant 10 mines 9x10", "Intermediaire 40 mines 18x14", "Avance 99 mines 30x16"};
		String rep = (String)JOptionPane.showInputDialog(null, "Selectionnez un niveau", "Saisie du niveau", JOptionPane.QUESTION_MESSAGE, null, niveaux, "Debutant 10 mines 9x10");
 
		Demineur d;
		if (rep.equals(niveaux[0])) {
			d= new Demineur(new TerrainMine(9,10,10));
		} else if (rep.equals(niveaux[1])) {
				d= new Demineur(new TerrainMine(18,14,40));
		}else {
				d= new Demineur(new TerrainMine(30,16,99));
		}
		// OU si vous avez la version 7 du JDK
		//switch (rep) {
		//   case "Debutant 10 mines 9x10" : d= new Demineur(new TerrainMine(9,10,10));break;
		//   case "Intermediaire 40 mines 18x14" : d= new Demineur(new TerrainMine(18,14,40));break;
		//   case "Avance 99 mines 30x16" : d= new Demineur(new TerrainMine(30,16,99));break;
		//   default : d= new Demineur(new TerrainMine(9,10,10));
		//}
		d.setVisible(true);
	}
}
 
class EcouteurBouton implements MouseListener{
	private ICase c;
	private Demineur d;
 
	public EcouteurBouton(Demineur d, ICase c) {
		this.c = c;
		this.d = d;
	}
	public void mouseClicked(MouseEvent e) {
			if (e.getButton()==MouseEvent.BUTTON1) { //
				if (this.getCase().estCachee() && !this.c.estMarquee()) {
				   this.getCase().clic();
				}
			} else {
				this.c.setMarquee(!this.c.estMarquee());
			}
			this.getDemineur().miseAJourImages();
			if (this.getDemineur().getTerrain().gagne()) {
				JOptionPane.showMessageDialog(null,  "BRAVO !", "Bravo, vous gagnez !", JOptionPane.PLAIN_MESSAGE);
				System.exit(0);
			}
			if (this.getDemineur().getTerrain().perdu()) {
				JOptionPane.showMessageDialog(null,  "Helas, vous perdez...", "Boummmm !", JOptionPane.PLAIN_MESSAGE);
				System.exit(0);
			}
	}
	public ICase getCase() {
		return this.c;
	}
	public Demineur getDemineur() {
		return this.d;
	}
	public void mousePressed(MouseEvent e) { }
	public void mouseReleased(MouseEvent e) { }
	public void mouseEntered(MouseEvent e) { }
	public void mouseExited(MouseEvent e) { }
}
